/*
 * Magic Eye - field strength indicator for TDA7021 FM radio
 *
 * 2.7 V => all LEDs off, 0.3 V => all LEDs on
 *
 * The algorithm takes the average of 4 ADC measurement and drives a 7 LED array. It maps 
 * 14 voltage steps to 7 LEDs. Even step-values are mapped to half the number of LEDs.
 * Odd values are mapped to half the number LEDs + the next LED at 25% (1 out of 4 cycles on) 
 *
 * Port A, Bit 0 ... 6: LED array, output
 * Port A, Bit 7: ADC, input
 *
 * Fuses: internal oscillator at 1 MHz, div 128 off 
 *
 * Created: 23.04.2013 21:11:17
 *  Author: Stephan Laage-Witt
 */ 

#define F_CPU 128000

#define ADC_UPPER_LIMIT 231				// corresponds to 2,70 V (with reference of 3.0 V)

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>

volatile uint16_t adc_result;
volatile uint8_t job_counter = 0;

/* Interrupt Service Handler ---------------------------------------------------------------------*/
/* Runs approx. 240 times per second. Takes ADC value. Counts job_counter through 0 to 3	      */ 

ISR(TIM0_COMPA_vect) {
	adc_result += ADCH;						// read ADC value, use upper 8 bits only
	ADCSRA |= (1<<ADSC);					// kick off new conversion 
	if (++job_counter == 4)					// increase job_counter
		job_counter = 0;
}	

/* MAIN ------------------------------------------------------------------------------------------*/
int main(void)
{
	const uint8_t led_array[8] = {	0b01111111,			// 7 LED array, 0 = on
									0b01111110,			// highest bit always 0
									0b01111100, 
									0b01111000, 
									0b01110000,
									0b01100000,
									0b01000000,
									0b00000000,
									};
	uint8_t step = 0;
	uint8_t odd_flag = 0;
									
	// I/O ports 
	DDRA = 0b01111111;			// bits 0 to 6 drive LED array, bit 7 used for ADC input

	// Initializing ADC 
	ADMUX  = (1<<MUX2) | (1<<MUX1) | (1<<MUX0);		// Vcc is reference, channel PDA7				
	ADCSRA = (1<<ADEN) | (1<<ADPS0);				// ADC enabled, pre-scaler division factor = 2
	ADCSRB = (1<<ADLAR);							// result is left adjusted in ADCH	
	ADCSRA |= (1<<ADSC);							// kick off first conversion 

	// Configuring timer 0 for interrupt at (approx) 240 Hz
	TCCR0A |= (1<<WGM01);							// CTC mode
	TCCR0B |= (1<<CS02) | (0<<CS01) | (0<<CS00);	// prescale = clock rate / 256 
	OCR0A = (16 - 1);								// compare value = 16 * 256 usec -> 244 Hz
	TIMSK0 |= (1<<OCIE0A);							// enable timer compare interrupt
	sei();											// off we go ...
	
	// set sleep mode to idle: interrupts and adc keep running 
	set_sleep_mode(SLEEP_MODE_IDLE);
	
    while(1) {
		if (job_counter == 3) {						// every 4th cycle:
			adc_result = adc_result / 4;			// take the average across 4 measurements
			if (adc_result > ADC_UPPER_LIMIT) {		// anything above upper limit (2.7 V) counts as "all LEDs off"
				step = 0;
			} else {				
				step = ADC_UPPER_LIMIT - adc_result;		// invert value, the lower the voltage the more LDEs on
				step /= 15;								// one step is approx. 0.18 V
				if (step > 14) step = 14;				// limit upper value to 14, minimum voltage approx. 0.3 V
				if (step & 0b00000001)					// check whether odd or even 
					odd_flag = 1;	
				else 
					odd_flag = 0;
				step = step / 2;					// reduce scale to 0 ... 7
				adc_result = 0;						// reset adc_result for the next round
			};			
		};
		
		if ((job_counter == 0) && (odd_flag))		// only at 1 out of 4 cycles:
			PORTA	= led_array[step + 1];			// if result is odd-number then toggle highest LED between on and off
		else	
			PORTA = led_array[step];				// set LED array according to ADC value

        sleep_mode();								// good night - wake up via interrupt				
	}
}	